#include "i0mmtopobarchart.h"
#include <qwt_plot_multi_barchart.h>
#include <qwt_legend.h>
#include <qwt_column_symbol.h>
#include <qwt_plot_grid.h>
#include <qwt_plot_canvas.h>
#include <qwt_scale_engine.h>
#include <qwt_text.h>
#include <qwt_plot_textlabel.h>
#include <qwt_scale_widget.h>
#include <QtDebug>
#include <qustring.h>

double LogTransform::bounded(double v) const {
    return qBound(0.01, v, LogMax);
}

QwtText NamesAxisDraw::label(double value) const {
    if(value < m_names.size()) {
        return m_names[value];
    }
    return QString::number(value);
}

void NamesAxisDraw::setNames(const QStringList &nl) {
    m_names = nl;
    invalidateCache();
}

I0MMTopoBarChart::I0MMTopoBarChart(QWidget *parent) : QwtPlot(parent)
{
    m_colormap[0] = "#ff5555"; m_colormap[1] = "#c8ff00"; m_colormap[2] = "#2882ff"; m_colormap[3] = "#7fffff" ;
    m_lambda_data_size = 0; // number of lambdas deduced from live data

    setAutoFillBackground( true );

    setPalette( Qt::white );
    canvas()->setPalette( QColor( "white" ) );
    setAxisTitle( QwtPlot::xBottom, "Intensity (μJ)" );

    m_min = m_max = 0;
    m_barchart = new QwtPlotMultiBarChart("Fel Intensity");
    m_barchart->setLayoutPolicy( QwtPlotMultiBarChart::AutoAdjustSamples );
    m_barchart->setSpacing(5);
    m_barchart->setMargin(5);
    m_barchart->setStyle( QwtPlotMultiBarChart::Grouped );
    m_barchart->setOrientation( Qt::Horizontal );
    m_barchart->attach( this );
    m_barchart->setLegendIconSize(QSize( 8, 10 ));
    QwtPlotGrid *grid = new QwtPlotGrid();
    grid->setPen(QColor(Qt::gray), 0.0, Qt::DashLine);
    grid->attach(this);

    NamesAxisDraw *yaxisdra = new NamesAxisDraw();
    this->setAxisScaleDraw(QwtPlot::yLeft, yaxisdra);
    axisScaleEngine(QwtPlot::yLeft)->setAttribute(QwtScaleEngine::Inverted, true);
    yaxisdra->setLabelAlignment(Qt::AlignLeft);

    // click on canvas
    canvas()->installEventFilter(new I0MapMonObjFilter(this));

    insertLegend( new QwtLegend() );

    setAutoReplot( true );
}

bool I0MMTopoBarChart::isEmpty() const {
    return m_barchart->barTitles().isEmpty();
}

bool I0MMTopoBarChart::stackedMode() const {
    return m_barchart->style() == QwtPlotMultiBarChart::Stacked;
}

QVector<int> I0MMTopoBarChart::lambdas() const {
    return m_cur_lambdas;
}

QwtPlotItem *I0MMTopoBarChart::findItem(const int rtti) const
{
    const QwtPlotItemList& ili = itemList(rtti);
    if(ili.size() > 0)
        return ili.first();
    return nullptr;
}

void I0MMTopoBarChart::update(const QList<CuData> &dl) {
    bool txf = axisTitle( QwtPlot::xBottom).text() == "Transmission";
    std::string error;
    // the cost of resetting m_min and m_max to -1 is small
    // but prevents malfunctioning when txf 0-1 data arrives late
    m_min = m_max = 0;
    QMap<int, double> valuemap; // <lambda-k - value> map
    QMap<int, double> lambda_item0_map; // <lambda-k item 0 - value> map
    if(dl.size() == static_cast<NamesAxisDraw *> (axisScaleDraw(QwtPlot::yLeft) )->size()) {
        QVector< QVector<double> > series;
        for(int i = 0; i < dl.size(); i++) {
            const CuData& d = dl[i];
            if(d["err"].toBool() ) {
                error += d["msg"].toString();
                QVector<double> errv;
                for(int i = 0; i < m_cur_lambdas.size(); i++)
                    errv << 0.0;
                series += errv;
            }
            else if(d["q"].toInt() == 1)
                ; // error += (QuString(d, "src").section('/', -1) + ":\ninvalid data quality\n").toStdString();
            else {
                std::vector<double> v = d["value"].toDoubleVector();
                // select lambdas
                for(size_t i = 0; i < v.size(); i++) {
                    if(m_cur_lambdas.contains(i)) {
                        if(v[i] < 0)
                            v[i] = 0;
                        if(txf && !lambda_item0_map.contains(i))
                            lambda_item0_map[i] = v[i];
                        txf ? valuemap[i] = v[i] / lambda_item0_map[i] : valuemap[i] = v[i];
                        if(m_min == m_max && m_min < 0) {
                            m_min = m_max = valuemap[i];
                        }
                        else {
                            if(m_min > valuemap[i]) m_min = valuemap[i];
                            if(m_max < valuemap[i]) m_max = valuemap[i];
                        }
                    }
                }
                series += valuemap.values().toVector();
            }
        }
        m_barchart->setSamples(series);

        const QVector<int> &lambdas = valuemap.keys().toVector();
        qDebug() << __PRETTY_FUNCTION__ << "size of data: " << dl.size() << "lambdas " << lambdas << " cur lambdas " << m_cur_lambdas;

        // removed: for some time after lambda changes lambdas.size may differ from m_cur_lambdas
        // do not call m_update_bars in this case
        //        if(lambdas.size() && m_cur_lambdas != lambdas) // if lambdas is empty it means error
        //            m_update_bars(lambdas);
        m_error(error); // show or hide error message
        setToolTip(error.c_str());
        setEnabled(error.length() == 0);

        //        if(isEnabled()) {
        if(m_max == m_min) m_max = 1.0;

        m_cur_data = dl; // save data
        m_max *= 1.05;

        if(m_min <= 0 && m_barchart->baseline() > 0) m_min= 0.01; // log scale
        if(axisScaleDiv(QwtPlot::xBottom).lowerBound() != m_min || axisScaleDiv(QwtPlot::xBottom).upperBound() != m_max)
            setAxisScale(QwtPlot::xBottom, m_min, m_max);
        //        }
    }
}

void I0MMTopoBarChart::update(const CuData &)
{

}

void I0MMTopoBarChart::clear() {
    m_barchart->setBarTitles(QList<QwtText>());
}

void I0MMTopoBarChart::setNames(const QStringList &names) {
    static_cast<NamesAxisDraw *> (axisScaleDraw(QwtPlot::yLeft) )->setNames(names);
    setAxisMaxMajor(QwtPlot::yLeft, names.size());
    setAxisMaxMinor(QwtPlot::yLeft, 0);
    setAxisScale(QwtPlot::yLeft,names.size() - 1, 0 );
}

void I0MMTopoBarChart::setLabels(const QStringList &ls) {
    if(!axisEnabled(QwtPlot::yRight)) {
        enableAxis(QwtPlot::yRight, true);
        NamesAxisDraw *y2axisdra = new NamesAxisDraw();
        setAxisScaleDraw(QwtPlot::yRight, y2axisdra);
        //        axisScaleEngine(QwtPlot::yRight)->setAttribute(QwtScaleEngine::Inverted, true);
        y2axisdra->setLabelAlignment(Qt::AlignRight);
    }
    static_cast<NamesAxisDraw *>(axisScaleDraw(QwtPlot::yRight))->setNames(ls);
    setAxisScale(QwtPlot::yRight, ls.size() - 1, 0);
    setAxisMaxMajor(QwtPlot::yRight, ls.size());
    setAxisMaxMinor(QwtPlot::yRight, 0);
    updateAxes();
    axisWidget(QwtPlot::yRight)->update();
}

void I0MMTopoBarChart::enableLogScale(bool en) {
    QwtScaleEngine *se;
    en ? se = new QwtLogScaleEngine : se = new QwtLinearScaleEngine;
    if(en) {
        se->setTransformation(new LogTransform);
        if(axisScaleDiv(QwtPlot::xBottom).lowerBound() < 0.01)
            setAxisScale(QwtPlot::xBottom, 0.01, m_max);
        m_barchart->setBaseline(0.01);
    }
    else{
        m_barchart->setBaseline(0.0);
        setAxisScale(QwtPlot::xBottom, m_min, m_max);
    }
    setAxisScaleEngine(QwtPlot::xBottom, se);
    replot();
}

void I0MMTopoBarChart::setLambdas(const QVector<int> lidxs) {
    qDebug() << __PRETTY_FUNCTION__ << lidxs;
    m_update_bars(lidxs);
    if(m_cur_data.size() > 0)
        update(m_cur_data);
    replot();
}

void I0MMTopoBarChart::setTxf(bool ) {
    m_min = m_max = -1;
}

void I0MMTopoBarChart::m_onBarsModeChanged(int mode) {
    mode == 0 ? m_barchart->setStyle( QwtPlotMultiBarChart::Stacked ) : m_barchart->setStyle( QwtPlotMultiBarChart::Grouped );
}

void I0MMTopoBarChart::m_update_bars(const QVector<int> &lambdas)
{
    qDebug() << __PRETTY_FUNCTION__ << lambdas;
    int idx = 0;
    m_barchart->resetSymbolMap();
    QList<QwtText> titles;
    foreach(int lambda, lambdas) {
        titles += QString("λ %1").arg(lambda + 1); // lambda symbol
        QwtColumnSymbol *symbol = new QwtColumnSymbol( QwtColumnSymbol::Box );
        symbol->setLineWidth( 2 );
        symbol->setFrameStyle( QwtColumnSymbol::Raised );
        if(m_colormap.contains(lambda))
            symbol->setPalette(QColor(m_colormap[lambda]));
        m_barchart->setSymbol( idx++, symbol );
    }
    m_barchart->setBarTitles(titles);
    updateLegend();

    // save current selected lambdas
    m_cur_lambdas = lambdas;
}

void I0MMTopoBarChart::m_error(const std::string &error)
{
    QwtPlotItem *ite = findItem(QwtPlotItem::Rtti_PlotTextLabel);
    if(error.length() > 0) {
        if(!ite) {
            ite = new QwtPlotTextLabel();
            ite->setZ(100);
            ite->attach(this);
        }
        QString s(error.c_str());
        s.remove("CuTangoWorld.extractData: attribute quality invalid -");
        s.replace(":", ":\n");
        static_cast<QwtPlotTextLabel *>(ite)->setText(QwtText(s));
    }
    else if(ite){
        ite->detach();
        delete ite;
    }
}
